@@ -1,14 +1,13 @@
 /*
- *  $Id: ui_getc.c,v 1.27 2005/11/27 16:38:28 tom Exp $
+ *  $Id: ui_getc.c,v 1.40 2007/07/04 12:43:54 tom Exp $
  *
- *  ui_getc.c - user interface glue for getc()
+ * ui_getc.c - user interface glue for getc()
  *
- * Copyright 2001-2004,2005	Thomas E. Dickey
+ * Copyright 2001-2006,2007 Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU Lesser General Public License as
- *  published by the Free Software Foundation; either version 2.1 of the
- *  License, or (at your option) any later version.
+ *  it under the terms of the GNU Lesser General Public License, version 2.1
+ *  as published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful, but
  *  WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -25,6 +24,10 @@
 #include <dialog.h>
 #include <dlg_keys.h>
 
+#ifdef NEED_WCHAR_H
+#include <wchar.h>
+#endif
+
 #if TIME_WITH_SYS_TIME
 # include <sys/time.h>
 # include <time.h>
@@ -172,6 +175,25 @@ dlg_flush_getc(void)
 #endif
 }
 
+/*
+ * Check if the stream has been unexpectedly closed, returning false in that
+ * case.
+ */
+static bool
+valid_file(FILE *fp)
+{
+    bool code = FALSE;
+    int fd = fileno(fp);
+
+    if (fd >= 0) {
+	long result = 0;
+	if ((result = fcntl(fd, F_GETFL, 0)) >= 0) {
+	    code = TRUE;
+	}
+    }
+    return code;
+}
+
 /*
  * Read a character from the given window.  Handle repainting here (to simplify
  * things in the calling application).  Also, if input-callback(s) are set up,
@@ -183,8 +205,10 @@ dlg_getc(WINDOW *win, int *fkey)
 {
     WINDOW *save_win = win;
     int ch = ERR;
+    int before_lookup;
     int result;
     bool done = FALSE;
+    bool literal = FALSE;
     DIALOG_CALLBACK *p;
     int interval = dialog_vars.timeout_secs;
     time_t expired = time((time_t *) 0) + dialog_vars.timeout_secs;
@@ -243,10 +267,25 @@ dlg_getc(WINDOW *win, int *fkey)
 	last_getc = ch;
 	*fkey = (ch > KEY_MIN && ch < KEY_MAX);
 #endif
+	if (literal) {
+	    done = TRUE;
+	    continue;
+	}
+
+	before_lookup = ch;
 	ch = dlg_lookup_key(win, ch, fkey);
+	dlg_trace_chr(ch, *fkey);
+
 	current = time((time_t *) 0);
 
 	switch (ch) {
+	case CHR_LITERAL:
+	    if (!literal) {
+		literal = TRUE;
+		keypad(win, FALSE);
+		continue;
+	    }
+	    break;
 	case CHR_REPAINT:
 	    (void) touchwin(win);
 	    (void) wrefresh(curscr);
@@ -261,14 +300,22 @@ dlg_getc(WINDOW *win, int *fkey)
 	    } else {
 		done = (interval <= 0);
 	    }
+	    if (!valid_file(stdin)
+		&& !valid_file(dialog_state.screen_output)) {
+		ch = ESC;
+		done = TRUE;
+	    }
 	    break;
+	case DLGK_FIELD_NEXT:
+	    /* FALLTHRU */
 	case TAB:
 	    /* Handle tab as a special case for traversing between the nominal
 	     * "current" window, and other windows having callbacks.  If the
 	     * nominal (control) window closes, we'll close the windows with
 	     * callbacks.
 	     */
-	    if (dialog_state.getc_callbacks != 0) {
+	    if (dialog_state.getc_callbacks != 0 &&
+		before_lookup == TAB) {
 		if ((p = dialog_state.getc_redirect) != 0) {
 		    p = p->next;
 		} else {
@@ -290,12 +337,26 @@ dlg_getc(WINDOW *win, int *fkey)
 		    dialog_state.getc_redirect = 0;
 		    win = save_win;
 		}
-		break;
 	    } else {
 		done = TRUE;
 	    }
+	    break;
+#ifdef NO_LEAKS
+	case DLG_CTRL('P'):	/* for testing, ^P closes the connection */
+	    close(0);
+	    close(1);
+	    close(2);
+	    break;
+#endif
+#ifdef HAVE_DLG_TRACE
+	case CHR_TRACE:
+	    dlg_trace_win(win);
+	    break;
+#endif
 	}
     }
+    if (literal)
+	keypad(win, TRUE);
     return ch;
 }
 
@@ -374,6 +435,7 @@ dlg_killall_bg(int *retval)
 			(void) signal(SIGHUP, finish_bg);
 		    (void) signal(SIGINT, finish_bg);
 		    (void) signal(SIGQUIT, finish_bg);
+		    (void) signal(SIGSEGV, finish_bg);
 		    while (dialog_state.getc_callbacks != 0) {
 			int fkey = 0;
 			dlg_getc_callbacks(ERR, fkey, retval);
